#!/usr/bin/env python
#-*- coding: utf-8 -*-
# vim:ts=4:sw=4:expandtab
#
# The MIT License (MIT)
# 
# Copyright (c) 2013 Lukas Märdian <maerdian@in.tum.de>
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import srf02, TokenRing
import alsaaudio, audioop
import collections, sys, signal, itertools, time

class sliceable_deque(collections.deque):
    def __getitem__(self, index):
        if isinstance(index, slice):
            return type(self)(itertools.islice(self,index.start,index.stop, index.step))
        return collections.deque.__getitem__(self, index)

#---------- High Resolution POSIX Timing Functions ----------#
import ctypes, os
CLOCK_MONOTONIC = 1 # see <linux/time.h>
RELATIVE = 0
ABSOLUTE = 1

class timespec(ctypes.Structure):
    _fields_ = [
        ('tv_sec', ctypes.c_long),
        ('tv_nsec', ctypes.c_long)
    ]

librt = ctypes.CDLL('librt.so.1', use_errno=True)

clock_nanosleep = librt.clock_nanosleep
clock_nanosleep.argtypes = [ctypes.c_int, ctypes.c_int, ctypes.POINTER(timespec), ctypes.POINTER(timespec)]
def nanosleep(sec,nsec,flags):
    request = timespec()
    request.tv_sec = sec
    request.tv_nsec = nsec
    remain = timespec()
    if clock_nanosleep(CLOCK_MONOTONIC, flags, ctypes.pointer(request), ctypes.pointer(remain)) != 0:
        errno_ = ctypes.get_errno()
        raise OSError(errno_, os.strerror(errno_))
    return remain.tv_sec + remain.tv_nsec * 1e-9

clock_gettime = librt.clock_gettime
clock_gettime.argtypes = [ctypes.c_int, ctypes.POINTER(timespec)]
def monotime():
    t = timespec()
    if clock_gettime(CLOCK_MONOTONIC, ctypes.pointer(t)) != 0:
        errno_ = ctypes.get_errno()
        raise OSError(errno_, os.strerror(errno_))
    return (t.tv_sec,t.tv_nsec)

def monotonic_time():
    """ Helper to return time as float instead of tuple """
    sec,nsec = monotime()
    return sec + nsec * 1e-9
#------------------------------------------------------------#

def h(signum, frame):
    print "now"


if __name__ == '__main__':
    try:
        ID = int(sys.argv[1])
    except IndexError:
        ID = None
    try:
        NUM = int(sys.argv[2])
    except IndexError:
        NUM = None
    try:
        DEV = str(sys.argv[3])
    except IndexError:
        DEV = '/dev/ttyUSB0'

    if len(sys.argv) == 1 or ID==None or NUM==None:
        print "Usage: %s ID NUM_AGENTS [DEVICE]"
        sys.exit(1)

    tr = TokenRing.TokenRing(DEV)
    tr.ID = ID
    tr.AGENTS = NUM
    # Autotune all sensors
    tr.calibrate()

    # Create new csv file
    title = "data_%d.csv" % ID
    f = open(title, 'w')
    line = "timestamp; recv/send box.sensor; (min) distance; min sector (sensor); raw measurements; debug"
    f.write(line+"\n")

inp = alsaaudio.PCM(alsaaudio.PCM_CAPTURE,alsaaudio.PCM_NORMAL)
last = monotonic_time()

q_len = (6*NUM)-6+2
q = sliceable_deque(maxlen=q_len)

# Set attributes: Mono, 48000 Hz, 16 bit little endian samples
inp.setchannels(1)
inp.setrate(48000)
inp.setformat(alsaaudio.PCM_FORMAT_S16_LE)
periodsize = 8
inp.setperiodsize(periodsize)

synced = False
first = True
INTERVAL = 0.2

high = [0,0,xrange(32500,33000),xrange(23000,33000),xrange(18000,33000)]
low  = [0,0,xrange(5000,32500), xrange(5000,23000), xrange(5000,18000)]

while True:
    s,n = monotime()
    now = s + n * 1e-9
    d = now-last
    # Read data from device
    l,data = inp.read()
    if l==periodsize:
        val = audioop.max(data,2)
        maximum = audioop.findmax(data,1)
    else:
        val = 0

    if synced and ((val >= 5000 and d >= INTERVAL-0.01) ):#or d>=INTERVAL):
        #avoid big delay, by continuing at now+5msec, using ABSOLUTE nanosleep
        n+=5000000
        if n >= 1000000000:
            s += 1
            n -= 1000000000
        if d >= INTERVAL*5-0.01:
            tr.handler(-1,-1)
            tr.handler(-1,-1)
            tr.handler(-1,-1)
            tr.handler(-1,-1)
        elif d >= INTERVAL*4-0.01:
            tr.handler(-1,-1)
            tr.handler(-1,-1)
            tr.handler(-1,-1)
        elif d >= INTERVAL*3-0.01:
            tr.handler(-1,-1)
            tr.handler(-1,-1)
        elif d >= INTERVAL*2-0.01:
            tr.handler(-1,-1)
        nanosleep(s,n,ABSOLUTE)
        tr.handler(maximum,0)
        last = now
    elif val >= 5000 and d >= INTERVAL-0.01:
        if d >= INTERVAL*5-0.01:
            q.extend([-1,-1,-1,-1])
        elif d >= INTERVAL*4-0.01:
            q.extend([-1,-1,-1])
        elif d >= INTERVAL*3-0.01:
            q.extend([-1,-1])
        elif d >= INTERVAL*2-0.01:
            q.append(-1)

        if first and val in high[NUM]:
            line = "%.6f START" % time.time()
            f.write(line+"\n")
            f.close()
            first = False
            print "Syncronizing..."
    
        if val in high[NUM]:
            q.append(1)
        elif val in low[NUM]:
            q.append(0)

        if len(q)==q_len and q[0]+q[q_len-1]==2 and not (1 in q[1:q_len-1]):
            synced = True
    
        #DEBUG the sync process
        #print q,val
        last = now
