#!/usr/bin/env python
#-*- coding: utf-8 -*-
# vim:ts=4:sw=4:expandtab
#
# The MIT License (MIT)
# 
# Copyright (c) 2013 Lukas Märdian <maerdian@in.tum.de>
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import sys, os, time, signal, struct, math, collections, thread
import srf02

#---------- High Resolution POSIX Timing Functions ----------#
import ctypes, os
CLOCK_MONOTONIC = 1 # see <linux/time.h>
RELATIVE = 0
ABSOLUTE = 1

class timespec(ctypes.Structure):
    _fields_ = [
        ('tv_sec', ctypes.c_long),
        ('tv_nsec', ctypes.c_long)
    ]

librt = ctypes.CDLL('librt.so.1', use_errno=True)

clock_nanosleep = librt.clock_nanosleep
clock_nanosleep.argtypes = [ctypes.c_int, ctypes.c_int, ctypes.POINTER(timespec), ctypes.POINTER(timespec)]
def nanosleep(sec,nsec,flags):
    request = timespec()
    request.tv_sec = sec
    request.tv_nsec = nsec
    remain = timespec()
    if clock_nanosleep(CLOCK_MONOTONIC, flags, ctypes.pointer(request), ctypes.pointer(remain)) != 0:
        errno_ = ctypes.get_errno()
        raise OSError(errno_, os.strerror(errno_))
    return remain.tv_sec + remain.tv_nsec * 1e-9

clock_gettime = librt.clock_gettime
clock_gettime.argtypes = [ctypes.c_int, ctypes.POINTER(timespec)]
def monotime():
    t = timespec()
    if clock_gettime(CLOCK_MONOTONIC, ctypes.pointer(t)) != 0:
        errno_ = ctypes.get_errno()
        raise OSError(errno_, os.strerror(errno_))
    return (t.tv_sec,t.tv_nsec)

def monotonic_time():
    """ Helper to return time as float instead of tuple """
    sec,nsec = monotime()
    return sec + nsec * 1e-9
#------------------------------------------------------------#

class TokenRing:
    def __init__(self, port='/dev/ttyUSB0'):
        """Setup TokenRing, containing six SRF02 ultrasonic senors"""
        self.i = 0
        self.j = 0
        self.c = 0
        self.ID = 0
        self.AGENTS = 4
        self.speed = 343 #assuming 20°C
        self.s0 = srf02.SRF02('e0', port)
        self.s1 = srf02.SRF02('e2', port)
        self.s2 = srf02.SRF02('e4', port)
        self.s3 = srf02.SRF02('e6', port)
        self.s4 = srf02.SRF02('e8', port)
        self.s5 = srf02.SRF02('ea', port)
        self.sensors = [self.s0, self.s1, self.s2, self.s3, self.s4, self.s5]
        #to send with sensor x only:
        #self.sensors = [self.s0]
        self.res_out = []
        self.write_out_buffer = []
        self.t_irq = collections.deque(maxlen=2)
        self.INTERVAL = 0.2

    def write_out(self, buf, ID):
        title = "data_%d.csv" % ID
        f = open(title, 'a')
        for line in buf:
            f.write(line+"\n")
        f.close()

    def calibrate(self):
        for s in [self.s0, self.s1, self.s2, self.s3, self.s4, self.s5]:
            for i in range(0,5):
                print s.range_cm(),
                sys.stdout.flush()
                time.sleep(0.065)
            print " ",
        print ""

    def usec_to_cm(self, usec):
        """Convert micro-seconds to centimeters"""
        return (usec*self.speed/10000.0)

    def usec_to_mm(self, usec):
        """Convert micro-seconds to millimeters"""
        return (usec*self.speed/1000.0)

    def set_speed(self, temp=20):
        """Set new speed coefficient, depending on 'temp' in °C"""
        self.speed = 20.05*math.sqrt(temp + 273.17)
        #alternative formula to calculate the speed-of-sound,
        #assuming it to be linear between -20°C and 40°C
        #self.speed = 331.6 + 0.6*temp

    def calc_res(self, res, delta_t):
        """Helper function to format the ranging results"""
        if res == 0:
            return -1 #ERROR
        else:
            final_res = self.usec_to_cm(res-delta_t*1000000)
            if final_res < 0 or final_res > 600:
                #print "xxx",final_res, res, str(delta_t*1000000)
                return -1 #ERROR
            else:
                #print str(int(round(final_res))),"cm"
                return int(round(final_res))

    def minimum_sensor(self):
        i = -1
        curr = 1000000
        curr_idx = -1
        for e in self.res_out:
            i+=1
            if e < curr and e != -1:
                curr = e
                curr_idx = i
        return curr_idx

    def handler(self, signum, frame):
        """Handle periodic ranging interrupts"""
        # Set start time at first time this handler is called
        self.t_irq.append(time.time())
        if signum == -1 and frame == -1: #ERROR
            if self.i == self.ID:
                line = "%.6f; s%d.%d; %3d cm"
            else:
                line = "%.6f; r%d.%d; %3d cm"
            line_out = line % (0.0, self.i, self.j, -1)
            debug = "; s" # This audio-signal was not recognized (skipped)
            line_out = line_out+debug
            self.write_out_buffer.append(line_out)
            print line_out
            if self.i == self.ID:
                thread.start_new_thread(self.write_out,(self.write_out_buffer,self.ID))
                self.write_out_buffer = []

            self.i = (self.i+1) % self.AGENTS
            if self.i == 0:
                self.j = (self.j+1) % len(self.sensors)
            #self.j = (self.j+1) % len(self.sensors)
            #if self.j == 0:
            #    self.i = (self.i+1) % self.AGENTS
            return
            
        if self.i == self.ID:
            # send burst + catch echo
            # sleep 35ms to wait for all receivers to initiate + 10ms tolerance buffer
            nanosleep(0,35000000,RELATIVE) 
            self.sensors[self.j].range_usec()
            time.sleep(0.065)
            # divide by 2, as this is a echo
            echo_range = self.calc_res(self.sensors[self.j].get_result()/2.0, 0)

            line = "%.6f; s%d.%d; %3d cm; %2d; []"
            line_data = (time.time(), self.i, self.j, echo_range, self.j)
            line_out = line % line_data
            debug = "; "+str(signum)
            line_out = line_out+debug
            self.write_out_buffer.append(line_out)
            print line_out
            thread.start_new_thread(self.write_out, (self.write_out_buffer, self.ID))
            self.write_out_buffer = []
        else:
            t0 = time.time()
            self.s0.fake_range_usec()
            t1 = time.time()
            self.s1.fake_range_usec()
            t2 = time.time()
            self.s2.fake_range_usec()
            t3 = time.time()
            self.s3.fake_range_usec()
            t4 = time.time()
            self.s4.fake_range_usec()
            t5 = time.time()
            self.s5.fake_range_usec()
            time.sleep(0.065)
            t6 = t0+0.035+0.00073 #TODO: make systematic error offset temp dependent, here: 25cm/0,0343 = 730 usec

            #clear output buffer
            self.res_out = []
            self.res_out.append(self.calc_res(self.s0.get_result(),(t6-t0)))
            self.res_out.append(self.calc_res(self.s1.get_result(),(t6-t1)))
            self.res_out.append(self.calc_res(self.s2.get_result(),(t6-t2)))
            self.res_out.append(self.calc_res(self.s3.get_result(),(t6-t3)))
            self.res_out.append(self.calc_res(self.s4.get_result(),(t6-t4)))
            self.res_out.append(self.calc_res(self.s5.get_result(),(t6-t5)))

            #print self.res_out, "cm"
            #print "Min. sensor:", self.minimum_sensor()
            mini = self.minimum_sensor()
            mini_val = self.res_out[mini]
            line = "%.6f; r%d.%d; %3d cm; %2d; [%3d, %3d, %3d, %3d, %3d, %3d]"
            line_data = (time.time(), self.i, self.j, mini_val, mini, self.res_out[0], self.res_out[1], self.res_out[2], self.res_out[3], self.res_out[4], self.res_out[5])

            debug = "; "+str(signum)

            line_out = line % line_data
            line_out = line_out+debug
            self.write_out_buffer.append(line_out)
            print line_out

        self.i = (self.i+1) % self.AGENTS
        if self.i == 0:
            self.j = (self.j+1) % len(self.sensors)
        #self.j = (self.j+1) % len(self.sensors)
        #if self.j == 0:
        #    self.i = (self.i+1) % self.AGENTS
        #self.c += self.INTERVAL
