#!/usr/bin/env python
#-*- coding: utf-8 -*-
# vim:ts=4:sw=4:expandtab
#
# The MIT License (MIT)
# 
# Copyright (c) 2013 Lukas Märdian <maerdian@in.tum.de>
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# This class provides access to SRF02 ultrasonic sensors, which are connected
# via the USB-I2c adapter.

import serial, time, os, select

class SRF02:
    def __init__(self, addr='e0', port='/dev/ttyUSB0'):
        """Setup SRF02 ultrasonic sensor with I2C address 'addr' on port 'port'"""
        self.port = port
        self.addr = int('0x'+addr,16)
        self.ser = serial.Serial(self.port, 19200, 8, 'N', 2, timeout=0.1)

    def __del__(self):
        """Close serial port"""
        self.ser.close()

    def i2c_write(self, bytes):
        """Write 'bytes' to this sensor via I2C
        This assumes the USB-I2C adapter to be used"""
        try:
            self.ser.flushInput()
            num = '%02d' % len(bytes.split(' '))
            self.ser.write(bytearray.fromhex(u'55 '+str(hex(self.addr).lstrip('0x'))+' 00 '+num+' '+bytes))
            res = self.ser.read(1) #read status byte
            return int(res.encode('hex'),16)
        except (IOError, ValueError, os.error, select.error):
            return 0 # return 0 on error

    def i2c_read(self, num_bytes, register='02'):
        """Read 'bytes' from 'register' of this sensor via I2C
        This assumes the USB-I2C adapter to be used"""
        try:
            self.ser.flushInput()
            bytes = '%02d' % num_bytes
            self.ser.write(bytearray.fromhex(u'55 '+str(hex(self.addr+1).lstrip('0x'))+' '+register+' '+bytes)) #writing to addr+1 == reading from addr
            res = self.ser.read(num_bytes) #read result
            return int(res.encode('hex'),16)
        except (IOError, ValueError, os.error, select.error):
            return 0 # return 0 on error

    def get_result(self):
        """Read the 2 bytes ranging result, starting at register '02'"""
        return self.i2c_read(2, '02')

    def get_min_range(self):
        """Read the 2 bytes min ranging range, starting at register '04'"""
        return self.i2c_read(2, '04')

    def range_inch(self):
        """Real Ranging Mode - Result in inches"""
        return self.i2c_write('50')

    def range_cm(self):
        """Real Ranging Mode - Result in centimeters"""
        return self.i2c_write('51')

    def range_usec(self):
        """Real Ranging Mode - Result in micro-seconds"""
        return self.i2c_write('52')

    def fake_range_inch(self):
        """Fake Ranging Mode - Result in inches"""
        return self.i2c_write('56')

    def fake_range_cm(self):
        """Fake Ranging Mode - Result in centimeters"""
        return self.i2c_write('57')

    def fake_range_usec(self):
        """Fake Ranging Mode - Result in micro-seconds"""
        return self.i2c_write('58')

    def transmit_burst(self):
        """Transmit an 8 cycle 40khz burst - no ranging takes place"""
        return self.i2c_write('5c')

    def force_autotune(self):
        """Force Autotune Restart - same as power-up. You can ignore this command."""
        return self.i2c_write('60')

    def set_addr(self, addr):
        """ATTENTION: This is only needed to setup new sensors"""
        self.i2c_write('a0')
        self.i2c_write('aa')
        self.i2c_write('a5')
        self.i2c_write(addr)
